//------------------------------------------------------------------
//example call:  ./rttyMod "CQ TEST CQ TEST ABC123 K"
//example compile:  g++ radwav_baudot.cpp -o radwav_baudot
//http://www.radwav.com 12|16|2014
//
//  This software was written by RadWAV for the SkyPi kit.  Visit
//  http://www.radwav.com for more information.  Essentially SkyPi
//  is a software defined radio for the Raspberry Pi.  It is used
//  for amateur radio and experimentation.
//
//  This example will convert a text message into Baudot code for
//  RTTY transmission.  The text message is passed into this code
//  using a command line argument.  (See example call above)
//
//  RadWAV wrote this example in C++ because this language offers
//  helpful features.  For instance <string> makes working with
//  strings much easier than in C.  Also <vector> makes working with
//  arrays of data much simple than in C.
//
//  Since this software was entirely written by RadWAV, it follows
//  RadWAV coding standards.  We hope you will find this standard
//  makes it easy to follow our code and learn about software radio!
//
//  Copyright (C) 2014 RadWAV, http://www.radwav.com
//  ------------------------------------------------
//  This program is free software:  you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation, either version 3
//  of the license, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warrant of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public
//  License along with this program.  If not, see
//  <http://www.gnu.org/licenses>
//
//  Please send comments or suggestions using the software form
//  at www.radwav.com under downloads.  Thank You!
//------------------------------------------------------------------

#include <string.h>
#include <iostream>
#include <string>
#include <bitset>
#include <vector>
#include <stdio.h>
using namespace std;


bool changeMode(bool letters, int characterIn)
//-----------------------------------------------------------------
//  changeMode(letters, charcterIn)
//    This function determines if the baudot code mode must change
//
//  Baudot uses two groups of characters which are called letters
//  and figures.  changeMode determines if the present mode must
//  change before the new character is sent.  If the mode must
//  change, then changeMode returns TRUE.  Otherwise, changeMode
//  returns FALSE.
//
//  RadWAV.com - 09|02|2014
//----------------------------------------------------------------
{
  //step 1:  define the characters in each of the two groups
  char letterGroup[] = " QWERTYUIOPASDFGHJKLZXCVBNM";
  char figureGroup[] = " 1234567890-!&#'()/:;?,.";

  //step 2:  determine if the characterIn belongs to the letters
  if(strchr(letterGroup, characterIn) != NULL)
    {
      if(letters)
        return false; //do not change mode
      else
        return true;  //change mode
    }

  //step 3:  determine if the characterIn belongs to the figures
  if(strchr(figureGroup , characterIn) != NULL)
   {
     if(letters)
      return true;   //change mode
    else
      return false;  //do not change mode
   }
}  //end of changeMode()

//--------------------------------------------------------------
//the main routine is where the program begins executing:
int main(int argc, char *argv[])
{
  char c;
  int i;

  //make look-up table to map ascii value to a baudot character
  short int lut[128];
  lut[0] = 0b00000;
  lut[32] = 0b00100;
  lut[81] = 0b11101;
  lut[87] = 0b11001;
  lut[69] = 0b10000;
  lut[82] = 0b01010;
  lut[84] = 0b00001;
  lut[89] = 0b10101;
  lut[85] = 0b11100;
  lut[73] = 0b01100;
  lut[79] = 0b00011;
  lut[80] = 0b01101;
  lut[65] = 0b11000;
  lut[83] = 0b10100;
  lut[68] = 0b10010;
  lut[70] = 0b10110;
  lut[71] = 0b01011;
  lut[72] = 0b00101;
  lut[74] = 0b11010;
  lut[75] = 0b11110;
  lut[76] = 0b01001;
  lut[90] = 0b10001;
  lut[88] = 0b10111;
  lut[67] = 0b01110;
  lut[86] = 0b01111;
  lut[66] = 0b10011;
  lut[78] = 0b00110;
  lut[77] = 0b00111;
  lut[15] = 0b00010;
  lut[12] = 0b01000;
  lut[36] = 0b11011;
  lut[35] = 0b11111;
  lut[0] = 0b00000;
  lut[32] = 0b00100;
  lut[49] = 0b11101;
  lut[50] = 0b11001;
  lut[51] = 0b10000;
  lut[52] = 0b01010;
  lut[53] = 0b00001;
  lut[54] = 0b10101;
  lut[55] = 0b11100;
  lut[56] = 0b01100;
  lut[57] = 0b00011;
  lut[48] = 0b01101;
  lut[45] = 0b11000;
  lut[7] = 0b10100;
  lut[63] = 0b10010;
  lut[33] = 0b10110;
  lut[38] = 0b01011;
  lut[35] = 0b00101;
  lut[39] = 0b11010;
  lut[40] = 0b11110;
  lut[41] = 0b01001;
  lut[34] = 0b10001;
  lut[47] = 0b10111;
  lut[58] = 0b01110;
  lut[59] = 0b01111;
  lut[63] = 0b10011;
  lut[44] = 0b00110;
  lut[46] = 0b00111;
  lut[15] = 0b00010;
  lut[12] = 0b01000;
  lut[36] = 0b11011;
  lut[35] = 0b11111;

  int n;
  std::vector<int> txBits;  //store bits to be transmitted here
  string str = argv[1];     //read text message passed into function

  bool letters = true; //true if presently in baudot letter set, false = figures

  //add some dead carrier before first start bit
  for(i=0; i<16; i++)
    txBits.push_back(1);

  for(i=0; i<str.size(); i++)  //loop through each element in str
  {
    c = str.at(i);             //present character from str

    if(c == ' ')  //send extra shift symbols as optional diddles
    {
      if(letters)  //sending in baudot letters mode
      {
        txBits.push_back(0);  //start-bit

        txBits.push_back(1);  //shift to letters = '11111'
        txBits.push_back(1);
        txBits.push_back(1);
        txBits.push_back(1);
        txBits.push_back(1);

        txBits.push_back(1);  //stop bits
        txBits.push_back(1);

      }
     else  //sending in baudot figures mode
     {
        txBits.push_back(0);    //start-bit

        txBits.push_back(1);    //shift to figures = '11011'
        txBits.push_back(1);
        txBits.push_back(0);
        txBits.push_back(1);
        txBits.push_back(1);

        txBits.push_back(1);   //stop bits
        txBits.push_back(1);
     }
    }

    if(changeMode(letters, c))
    {
      letters = !letters;       //switch mode

      if(letters)               //changing to letters mode
      {
        txBits.push_back(0);    //start-bit

        txBits.push_back(1);    //shift to letters is '11111'
        txBits.push_back(1);
        txBits.push_back(1);
        txBits.push_back(1);
        txBits.push_back(1);

        txBits.push_back(1);    //send a few stop bits
        txBits.push_back(1);
      }
      else
      {
        txBits.push_back(0);    //start-bit

        txBits.push_back(1);    //shift to figures is '11011'
        txBits.push_back(1);
        txBits.push_back(0);
        txBits.push_back(1);
        txBits.push_back(1);

        txBits.push_back(1);    //send a few stop bits
        txBits.push_back(1);
      }
    }

    //now send the current character
    txBits.push_back(0);       //start-bit

    for (n=4; n>=0; n--)       //look-up the binary baudot code
    {
      if (((lut[c]) >> n) & 1)
        txBits.push_back(1);   //send a '1'
      else
        txBits.push_back(0);   //send a '0'
    }
    //add some stop bits
    txBits.push_back(1);
    txBits.push_back(1);
    txBits.push_back(1);

  }

  //print the encoded bits to stdout
  for(n=0; n<txBits.size(); n++)
    {
      printf("%i", txBits.at(n));
    }

  printf("\n");

  return(1);
}
