//-----------------------------------------------------------------------------
//  radwav_rtty is used to modulate amateur radio RTTY using the RadWAV SkyPi.
//  RTTY is the abbreviation for "Radio Teletype"
//
//  Example Call:
//   ./radwav_baudot "MY RADIO IS A HOMEBREW RADWAV" | xargs sudo ./radwav_rtty 7.080
//  This example will immediately send the message when you press enter.
//  Change the last number to vary the carrier frequency.  Also, see the RadWAV
//  Python RTTY example which simplifies the user interface.
//
//  To Compile:  gcc radwav_rtty12162014.c -std=c99 -lm -lwiringPi -o radwav_rtty
//
//  For more information, visit http://www.radwav.com
//
//  Normally this code is used along with radwav_baudot.cpp which encodes
//  text messages (with numbers also allowed) into the Baudot code.  This is
//  standard for RTTY.  Although, amateurs are free to use different codes
//  and add powerful Forward Error Correction codes provided they are published.
//
//  This program began with the work from:
//  http:// www.icrobotics.co.uk/wiki/index.php/Turning_the_Raspberry_Pi_Into_an_FM_Transmitter
//  Credit:  Oliver Mattos & Oskar Weigl who made the code available under GPL
//
//  RadWAV adapted the code for amateur radio RTTY transmission and specifically
//  for the SkyPi-40 software defined radio kit.
//
//  Copyright (c) 2014 RadWAV
//  -------------------------
//  This program is free software:  you can redistribute it and/or modify
//  it under thte terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that is will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.or/licenses/>.
//
//  If you have suggestions for this software, you may submit them at:
//  http://www.radwav.com using the form on the downloads page.  Thank You.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <math.h>
#include <fcntl.h>
#include <assert.h>
#include <malloc.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>

#include <wiringPi.h>

#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)

int  mem_fd;
char *gpio_mem, *gpio_map;
char *spi0_mem, *spi0_map;


// I/O access
volatile unsigned *gpio;
volatile unsigned *allof7e;

// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

#define GPIO_SET *(gpio+7)  // sets   bits which are 1 ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0
#define GPIO_GET *(gpio+13)  // sets   bits which are 1 ignores bits which are 0

#define ACCESS(base) *(volatile int*)((int)allof7e+base-0x7e000000)
#define SETBIT(base, bit) ACCESS(base) |= 1<<bit
#define CLRBIT(base, bit) ACCESS(base) &= ~(1<<bit)

#define CM_GP0CTL (0x7e101070)
#define GPFSEL0 (0x7E200000)
#define CM_GP0DIV (0x7e101074)
#define CLKBASE (0x7E101000)
#define DMABASE (0x7E007000)
#define PWMBASE  (0x7e20C000) /* PWM controller */


struct GPCTL {
    char SRC         : 4;
    char ENAB        : 1;
    char KILL        : 1;
    char             : 1;
    char BUSY        : 1;
    char FLIP        : 1;
    char MASH        : 2;
    unsigned int     : 13;
    char PASSWD      : 8;
};



void getRealMemPage(void** vAddr, void** pAddr) {
    void* a = valloc(4096);
    ((int*)a)[0] = 1;  // use page to force allocation.
    mlock(a, 4096);  // lock into ram.
    *vAddr = a;  // yay - we know the virtual address
    unsigned long long frameinfo;
    int fp = open("/proc/self/pagemap", 'r');
    lseek(fp, ((int)a)/4096*8, SEEK_SET);
    read(fp, &frameinfo, sizeof(frameinfo));
    *pAddr = (void*)((int)(frameinfo*4096));
}

void freeRealMemPage(void* vAddr) {
    munlock(vAddr, 4096);  // unlock ram.
    free(vAddr);
}

void setup_fm()
{

    /* open /dev/mem */
    if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
        printf("can't open /dev/mem \n");
        exit (-1);
    }
    allof7e = (unsigned *)mmap(
                  NULL,
                  0x01000000,  //len
                  PROT_READ|PROT_WRITE,
                  MAP_SHARED,
                  mem_fd,
                  0x20000000  //base
              );

    if ((int)allof7e==-1) exit(-1);

    SETBIT(GPFSEL0 , 14);
    CLRBIT(GPFSEL0 , 13);
    CLRBIT(GPFSEL0 , 12);

   //struct GPCTL setupword = {6/*SRC*/, 1, 0, 0, 0, 1,0x5a};
   //RadWAV Modification 10/21/2014
   struct GPCTL setupword = {6, 1, 0, 0, 0, 3, 0x5a};

   ACCESS(CM_GP0CTL) = *((int*)&setupword);
}


void modulate(int m)
{
    ACCESS(CM_GP0DIV) = (0x5a << 24) + 0x4d72 + m;
}

struct CB {
    volatile unsigned int TI;
    volatile unsigned int SOURCE_AD;
    volatile unsigned int DEST_AD;
    volatile unsigned int TXFR_LEN;
    volatile unsigned int STRIDE;
    volatile unsigned int NEXTCONBK;
    volatile unsigned int RES1;
    volatile unsigned int RES2;
};

struct DMAregs {
    volatile unsigned int CS;
    volatile unsigned int CONBLK_AD;
    volatile unsigned int TI;
    volatile unsigned int SOURCE_AD;
    volatile unsigned int DEST_AD;
    volatile unsigned int TXFR_LEN;
    volatile unsigned int STRIDE;
    volatile unsigned int NEXTCONBK;
    volatile unsigned int DEBUG;
};

struct PageInfo {
    void* p;  // physical address
    void* v;   // virtual address
};

struct PageInfo constPage;
struct PageInfo instrPage;
struct PageInfo instrs[1024];


void playWav(char dataIn[], float samplerate)
{
    int bufPtr=0;
    float datanew, dataold = 0;
    short data;
    float fmconstant = samplerate * 50.0e-6;  // for pre-emphisis filter.  50us time constant
    int clocksPerSample = 22500.0/samplerate*1400.0;  // for timing
    int symbolCounter = 0;
    float dval  = 0.0;

    while(symbolCounter < strlen(dataIn))  //RadWAV frequency shift keying modulation
      {
        if(dataIn[symbolCounter] ==  '1')      //send a 1
          dval = 7.02;
        else if(dataIn[symbolCounter] == '0')  //send a 0
          dval = 0.0;
        else
          printf("\nUnexpected value (other than 1 or 0) given in dataIn: %c\n", dataIn[symbolCounter]);

        symbolCounter++;

        for (float i=0; i<485; i+=1) {//RadWAV:  22050 sample rate / 45.45 symbol rate = 485.149

          //float sample = datanew + (dataold-datanew) / (1-fmconstant);  // fir of 1 + s tau
          //float dval = sample*15.0;  // actual transmitted sample.  15 is bandwidth (about 75 kHz)

          int intval = (int)(round(dval));  // integer component
          float frac = (dval - (float)intval)/2 + 0.5;
          unsigned int fracval = frac*clocksPerSample;

          bufPtr++;
          while( ACCESS(DMABASE + 0x04 /* CurBlock*/) ==  (int)(instrs[bufPtr].p)) usleep(1000);
          ((struct CB*)(instrs[bufPtr].v))->SOURCE_AD = (int)constPage.p + 2048 + intval*4 - 4 ;

          bufPtr++;
          while( ACCESS(DMABASE + 0x04 /* CurBlock*/) ==  (int)(instrs[bufPtr].p)) usleep(1000);
          ((struct CB*)(instrs[bufPtr].v))->TXFR_LEN = clocksPerSample-fracval;

          bufPtr++;
          while( ACCESS(DMABASE + 0x04 /* CurBlock*/) ==  (int)(instrs[bufPtr].p)) usleep(1000);
          ((struct CB*)(instrs[bufPtr].v))->SOURCE_AD = (int)constPage.p + 2048 + intval*4+4;

          bufPtr=(bufPtr+1) % (1024);
          while( ACCESS(DMABASE + 0x04 /* CurBlock*/) ==  (int)(instrs[bufPtr].p)) usleep(1000);
          ((struct CB*)(instrs[bufPtr].v))->TXFR_LEN = fracval;

          dataold = datanew;
      }  //end of for loop
     }  //end of  while loop
    //close(fp);
}

void unSetupDMA(){
    printf("exiting\n");
    struct DMAregs* DMA0 = (struct DMAregs*)&(ACCESS(DMABASE));
    DMA0->CS =1<<31;  // reset dma controller
}

void handSig() {
    exit(0);
}
void setupDMA( float centerFreq ){

  atexit(unSetupDMA);
  signal (SIGINT, handSig);
  signal (SIGTERM, handSig);
  signal (SIGHUP, handSig);
  signal (SIGQUIT, handSig);

   // allocate a few pages of ram
   getRealMemPage(&constPage.v, &constPage.p);
   int centerFreqDivider = (int)((500.0 / centerFreq) * (float)(1<<12) + 0.5);

   // make data page contents - it's essientially 1024 different commands for the
   // DMA controller to send to the clock module at the correct time.
   for (int i=0; i<1024; i++)
     ((int*)(constPage.v))[i] = (0x5a << 24) + centerFreqDivider - 512 + i;


   int instrCnt = 0;

   while (instrCnt<1024) {
     getRealMemPage(&instrPage.v, &instrPage.p);

     // make copy instructions
     struct CB* instr0= (struct CB*)instrPage.v;

     for (int i=0; i<4096/sizeof(struct CB); i++) {
       instrs[instrCnt].v = (void*)((int)instrPage.v + sizeof(struct CB)*i);
       instrs[instrCnt].p = (void*)((int)instrPage.p + sizeof(struct CB)*i);
       instr0->SOURCE_AD = (unsigned int)constPage.p+2048;
       instr0->DEST_AD = PWMBASE+0x18 /* FIF1 */;
       instr0->TXFR_LEN = 4;
       instr0->STRIDE = 0;
       //instr0->NEXTCONBK = (int)instrPage.p + sizeof(struct CB)*(i+1);
       instr0->TI = (1/* DREQ  */<<6) | (5 /* PWM */<<16) |  (1<<26/* no wide*/) ;
       instr0->RES1 = 0;
       instr0->RES2 = 0;

       if (i%2) {
         instr0->DEST_AD = CM_GP0DIV;
         instr0->STRIDE = 4;
         instr0->TI = (1<<26/* no wide*/) ;
       }

       if (instrCnt!=0) ((struct CB*)(instrs[instrCnt-1].v))->NEXTCONBK = (int)instrs[instrCnt].p;
       instr0++;
       instrCnt++;
     }
   }
   ((struct CB*)(instrs[1023].v))->NEXTCONBK = (int)instrs[0].p;

   // set up a clock for the PWM
   ACCESS(CLKBASE + 40*4 /*PWMCLK_CNTL*/) = 0x5A000026;
   usleep(1000);
   ACCESS(CLKBASE + 41*4 /*PWMCLK_DIV*/)  = 0x5A002800;
   ACCESS(CLKBASE + 40*4 /*PWMCLK_CNTL*/) = 0x5A000016;
   usleep(1000);

   // set up pwm
   ACCESS(PWMBASE + 0x0 /* CTRL*/) = 0;
   usleep(1000);
   ACCESS(PWMBASE + 0x4 /* status*/) = -1;  // clear errors
   usleep(1000);
   ACCESS(PWMBASE + 0x0 /* CTRL*/) = -1; //(1<<13 /* Use fifo */) | (1<<10 /* repeat */) | (1<<9 /* serializer */) | (1<<8 /* enable ch */) ;
   usleep(1000);
   ACCESS(PWMBASE + 0x8 /* DMAC*/) = (1<<31 /* DMA enable */) | 0x0707;

   //activate dma
   struct DMAregs* DMA0 = (struct DMAregs*)&(ACCESS(DMABASE));
   DMA0->CS =1<<31;  // reset
   DMA0->CONBLK_AD=0;
   DMA0->TI=0;
   DMA0->CONBLK_AD = (unsigned int)(instrPage.p);
   DMA0->CS =(1<<0)|(255 <<16);  // enable bit = 0, clear end flag = 1, prio=19-16
}

//======================================================================
//  The main routine follows - start here to understand the code:
//======================================================================

int main(int argc, char **argv)
{
   int i;                   //used to print out the arguments passed-in
   int keyPowerAmpPin = 5;  //GPIO 23 for RadWAV SkyPi
   int txEnaPin = 4;        //GPIO 24 for RadWAV SkyPi

   printf("\n[STATUS] - radwav_rtty version 12162014");
   printf("\n[radwav_rtty.c] Number of arguments:  %i", argc);
   for(i=0; i<argc; i++)
     printf("\nargument[%i]: %s", i, argv[i]);

    if (argc==3) {
      setup_fm();
      setupDMA(argc>2?atof(argv[1]):7.090);  //set-up carrier frequency

      printf("\n[STATUS] - Frequency = %.3f [MHz]", argc>2?atof(argv[1]):7.090);

      if(wiringPiSetup() == -1)
        {
         printf("\n[ERROR] - WiringPi setup failed.  Exiting.\n");
         return 1;
        }

      //set the RadWAV SkyPi control pins as outputs
      pinMode(keyPowerAmpPin, OUTPUT);
      pinMode(txEnaPin, OUTPUT);

      //key-up the RadWAV SkyPi power amplifier and the buffer
      digitalWrite(keyPowerAmpPin, 1);
      digitalWrite(txEnaPin, 1);

      //Baudot encoded message is stored in argv[2]
      playWav(argv[2], argc>3?atof(argv[3]):22500);

      //shut down the power amplifier and buffer
      digitalWrite(keyPowerAmpPin, 0);
      digitalWrite(txEnaPin, 0);

      //set the transmitter LO where it will not radiate or interfere
      setup_fm();
      setupDMA(20.0);  //SkyPi-40 low-pass filter will reject 20MHz

    } else
    {
     // fprintf(stderr, "Usage:   program wavfile.wav [freq] [sample rate]\n\nWhere wavfile is 16 bit 22.5kHz Mono.  Set wavfile to '-' to use stdin.\nfreq is in Mhz (default 103.3)\nsample rate of wav file in Hz\n\nPlay an empty file to transmit silence\n");
        fprintf(stderr, "Usage:   ./radwav_rtty [freq] 011010010001001010101");
        fprintf(stderr, "This may be used to send standard radio-teletype messages:");
        fprintf(stderr, "./radwav_baudot \"CQ CQ CQ DE ABCDEF\" | xargs sudo ./radwav_rtty 7.090");
    }
    printf("\n[STATUS] - Finished.");
    return 0;

} // main

