/*  This code has been adapted for the RADWAV SkyPi to set the synthesizer frequency.

    In order to compile this code, you will need to have wiringPi installed on the
    Raspberry Pi.  This is used to access hardware features of the RADWAV SkyPi.

    To complile:  gcc tunePll.c -std=c99 -lm -lwiringPi -o tuneRadWAV

    To run:  sudo ./tuneRadWAV 28.160 will place the transmitter near the 40 meter QRP
    calling frequency at 7.040 MHz.  The receiver divides the PLL frequency by four.

    The SkyPi provides a power-amplifier, low-pass filtering, and a linear regulator
    for +5V that is supplied to the Raspberry Pi over the GPIO connector.  There is also
    an optional recevier port.  This code is normally used with the optional skypirx.
    For more details on the hardware, please visit:
    http://www.radwav.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    The code is based on outstanding earlier works by major contributors
    including:

     - WSPR code from F8CHK
     -PiFM code from http://www.icrobotics.co.uk/wiki/index.php/Turning_the_Raspberry_Pi_Into_an_FM_Transmitter
     -And the work of Dan MD1CLV

    73 DE RADWAV (FEB 5, 2015)
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h>
#include <math.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <wiringPi.h>           //Added to control RADWAV radio hardware

// RF code:
#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO controller */
#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)

int  mem_fd;
char *gpio_mem, *gpio_map;
char *spi0_mem, *spi0_map;

// I/O access
volatile unsigned *gpio;
volatile unsigned *allof7e;

// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

#define GPIO_SET *(gpio+7)  // sets bits which are 1 ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0
#define GPIO_GET *(gpio+13) // sets bits which are 1 ignores bits which are 0

#define ACCESS(base) *(volatile int*)((int)allof7e+base-0x7e000000)
#define SETBIT(base, bit) ACCESS(base) |= 1<<bit
#define CLRBIT(base, bit) ACCESS(base) &= ~(1<<bit)
#define CM_GP0CTL (0x7e101070)
#define GPFSEL0 (0x7E200000)
#define CM_GP0DIV (0x7e101074)


struct GPCTL {
    char SRC         : 4;
    char ENAB        : 1;
    char KILL        : 1;
    char             : 1;
    char BUSY        : 1;
    char FLIP        : 1;
    char MASH        : 2;
    unsigned int     : 13;
    char PASSWD      : 8;
};

typedef struct tuning_data {
  double requested;
  unsigned long tuning_word;
  double actual;
  }  tuning_data;

void txon()
{

    allof7e = (unsigned *)mmap(
                  NULL,
                  0x01000000,  //len
                  PROT_READ|PROT_WRITE,
                  MAP_SHARED,
                  mem_fd,
                  0x20000000  //base
              );

    if ((int)allof7e==-1) exit(-1);

    SETBIT(GPFSEL0 , 14);
    CLRBIT(GPFSEL0 , 13);
    CLRBIT(GPFSEL0 , 12);

    //radwav changed to MASH2
    struct GPCTL setupword = {6/*SRC*/, 1, 0, 0, 0, 2,0x5a};
    ACCESS(CM_GP0CTL) = *((int*)&setupword);
}

void txoff()
{
    //radwav changed to MASH2
    struct GPCTL setupword = {6/*SRC*/, 0, 0, 0, 0, 2,0x5a}; 
    ACCESS(CM_GP0CTL) = *((int*)&setupword); 
}

void setfreq(long freq)
{
    ACCESS(CM_GP0DIV) = (0x5a << 24) + freq;
}

//
// Set up a memory regions to access GPIO
//
void setup_io()
{
    /* open /dev/mem */
    if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
        printf("can't open /dev/mem \n");
        exit (-1);
    }

    /* mmap GPIO */

    // Allocate MAP block
    if ((gpio_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1))) == NULL) {
        printf("allocation error \n");
        exit (-1);
    }

    // Make sure pointer is on 4K boundary
    if ((unsigned long)gpio_mem % PAGE_SIZE)
        gpio_mem += PAGE_SIZE - ((unsigned long)gpio_mem % PAGE_SIZE);

    // Now map it
    gpio_map = (unsigned char *)mmap(
                   gpio_mem,
                   BLOCK_SIZE,
                   PROT_READ|PROT_WRITE,
                   MAP_SHARED|MAP_FIXED,
                   mem_fd,
                   GPIO_BASE
               );

    if ((long)gpio_map < 0) {
        printf("mmap error %d\n", (int)gpio_map);
        exit (-1);
    }

    // Always use volatile pointer!
    gpio = (volatile unsigned *)gpio_map;


}

void setup_gpios()
{
   int g;

    // Set GPIO pins 7-11 to output
    for (g=7; g<=11; g++) {
        INP_GPIO(g); // must use INP_GPIO before we can use OUT_GPIO
        //OUT_GPIO(g);
    }

}


void calculate_tuning_info(tuning_data* tuning_info)
{
  double divisor;
  unsigned long decimal_part;
  unsigned long fractional_part;
  double actual_divisor;

  divisor = (double)500000000/tuning_info->requested;
  decimal_part = (unsigned long) divisor;
  fractional_part = (divisor - decimal_part) * (1 << 12);
  tuning_info->tuning_word = decimal_part * (1 << 12) + fractional_part;
  actual_divisor = (double)tuning_info->tuning_word / (float)(1 << 12);

  tuning_info->actual = (double)500000000 / actual_divisor;
}

/*void sym_to_tuning_words(double base_freq, unsigned char* wspr_symbols, unsigned long* tuning_words)
{
  int i;
  double symbol_freq;
  tuning_data tuning_info[4];


  for (i = 0; i < 4; i++)
  {
    symbol_freq = base_freq + (i-2) * WSPR_OFFSET;
    tuning_info[i].requested = symbol_freq;
    calculate_tuning_info(&tuning_info[i]);
    printf("Symbol %d: Target freq=%fHz, Actual freq=%fHz, Error=%fHz, Tuning Word=%lx\n", i, symbol_freq, tuning_info[i].actual, symbol_freq-tuning_info[i].actual, tuning_info[i].tuning_word);
  }

  for (i = 0; i < 162; i++)
  {
    tuning_words[i] = tuning_info[wspr_symbols[i]].tuning_word;
  }
}
*/


//============================================================================
//  The main routine follows.  This is where you normally might want to
//  modify the code to customize how your RadWAV SkyPi operates
//============================================================================

int main(int argc, char *argv[])
{
  unsigned long tuning_words[1];
  int i;
  double centre_freq, default_freq = 7040000;  //default to 7.040MHz
  double inactive_freq = 15000000;             //SkyPi-40 will not pass this
  int inactiveCounter = 0;

  //radwav.com additions--------------------------------------------------
  int keyPowerAmpPin = 5;    //GPIO24
  int txEnaPin = 4;          //GPIO23
  int morseCodeKeyPin = 0;   //GPIO17 - one of the lines on SkyPi P3

  //printf("\n[STATUS] - RADWAV_CW starting.  Press CTRL-C to end...");
  //printf("\n[INFO] - Version 12202014 - For use with RadWAV SkyPi-40.");

  if(wiringPiSetup() == -1)
  {
    printf("\n[ERROR] - WiringPi setup failed.  Exiting.\n");
  }

  if(argc != 2){
    printf("Usage: radwav_cw <frequency in MHz>\n");
    printf("\te.g.: radwav_cw 7.040000\n");
    printf("\nFor more details, visit http://www.radwav.com\n");
    return 1;
  }

  if(strcmp(argv[0], "./tuneRadWAV") != 0) {
    printf("\nUnrecognized command: %s", argv[0]);
    printf("\nFor more details, visit http://www.radwav.com\n");
    return 1;
  }

  centre_freq = atof(argv[1]) * 1.0E6;  //convert to units [Hertz]

  //check if transmit frequency is within a reasonble range
  if((centre_freq < 1.5E6) || (centre_freq > 50.0E6)) {
    printf("\n[WARNING] - Centre Frequency %f [MHz] is Out-Of-Range", centre_freq / 1E6);
    printf("\n[WARNING] - Centre Frequency will be %f [MHz]", default_freq / 1E6);
    centre_freq = default_freq;  //Hertz
  }

  printf("\n[STATUS] - Frequency %f [MHz]", centre_freq / 1E6);

  //set-up wiringPi
  setup_io();
  setup_gpios();
  txon();

  //set-up the pins for the RadWAV SkyPi transmitter hardware
  pinMode(keyPowerAmpPin, OUTPUT);  //enables envelope shaped power amplifier
  pinMode(txEnaPin, OUTPUT);        //enables buffer amplifier
  pinMode(morseCodeKeyPin, INPUT);  //connects to SkyPi P3 I/O Jack

  //set-up initial states for the power amplifier and buffer stage
  digitalWrite(keyPowerAmpPin, 0);  //power amplifier is off
  digitalWrite(txEnaPin, 0);        //buffer amplifier is off

  //printf("\n[STATUS] - Ready to Transmit.");
  //printf("\n[INFO] - Connect optional RadWAV I/O cable red & green wires to a key.\n");
  //printf("\n[INFO] - Close Key Connected to RADWAV SkyPi P3 Jack to Transmit...\n");
  printf("\n");

  //tune into the reject band when not sending
  //tuning_data tuning_info[1];
  //tuning_info[0].requested = inactive_freq;      
  //calculate_tuning_info(&tuning_info[0]);
  //tuning_words[0] = tuning_info[0].tuning_word;
  //setfreq(tuning_words[0]);

  //tune to the transmit frequency
  tuning_data tuning_info[1];  
  tuning_info[0].requested = centre_freq;      
  calculate_tuning_info(&tuning_info[0]);
  tuning_words[0] = tuning_info[0].tuning_word;
  setfreq(tuning_words[0]);

  //txoff();

  //unkey radwav transmitter hardware
  digitalWrite(keyPowerAmpPin, 0);
  digitalWrite(txEnaPin, 0);

  //printf("Done!\n");

  return 0;
}
